export interface Stringable {
    toString(): string;
}
export declare type Template<componentType extends Component<any>> = (component: componentType) => Stringable;
/**
 * A data structure representing one or more components.
 */
export interface DeflatedComponent {
    type: string;
    state: any;
    id: string;
    label: string;
    children?: DeflatedComponent[];
}
export interface ComponentClass<ComponentType extends Component<StateType>, StateType extends {}> {
    new (state: StateType): ComponentType;
    getIdentifier(label: string): string;
}
/**
 * The base for all Armature components.
 */
export declare class Component<StateType extends {}> {
    /**
     * A list of class names to apply to the component's root element.
     */
    static htmlClassNames: string[];
    /**
     * The HTML tag this component type will be using when reified.
     */
    static htmlTagName: string;
    /**
     * The template used to render this component.
     * Use @Template to specify the template for a component type.
     */
    template: Template<this>;
    /**
     * The state backing this component.
     */
    state: StateType;
    /**
     * The generated globally unique identifier for this component.
     */
    id: string;
    /**
     * The given label identifying this component.
     * This should be unique among all siblings of this component.
     */
    label: string;
    /**
     * The element representing this component, if one exists.
     */
    element: HTMLElement;
    /**
     * The components this component contains.
     */
    children: Component<any>[];
    /**
     * The component containing this component, if one exists.
     */
    parent: Component<any>;
    /**
     * A flag telling ekma that this object yields HTML
     */
    private safeInHTML;
    /**
     * Creates a new component with the given data.
     *
     * @param data The data to initialize the component with
     */
    constructor(state: StateType);
    /**
     * Deprecated: use Parent#getChild and Child#setParent instead.
     *
     * Creates or retrieves a component that is attached to another component.
     * Using the same parent and 'label' parameters will retrieve the same object.
     * The third parameter is passed directly to the constructor of a created component.
     *
     * @param parent The component to attach to
     * @param label The label to use when creating or retrieving the component
     * @param state The state to construct the component with if it hasn't been created.
     */
    static for(parent: Component<any>, label: string, state: any): Component<any>;
    /**
     * Returns the child component of the given type and label, if it exists.
     * Returns a Component<any>, requires casting to the type passed to the function.
     *
     * @param childClass The child's type to search for.
     * @param label The child's label to search for.
     */
    getChild(childClass: typeof Component, label: string): Component<any>;
    /**
     * Sets the label of the component and returns it.
     *
     * @param label The label to set.
     */
    setLabel(label: string): this;
    /**
     * Sets this component to be a child of the given component.
     * A component can only be the child of one component.
     * This method unsets an existing parent if one was set.
     *
     * @param parent The component to contain this one.
     */
    setParent(parent: Component<any>): this;
    /**
     * Returns the default state of the component if none is given at construction time.
     * Returning null from this function signifies that state must be specified.
     */
    getDefaultState(): StateType;
    setState(state: StateType): void;
    private $onChange;
    onChange(method: Function): void;
    /**
     * Serializes the component's data for potential recreation.
     */
    packState(): any;
    static unpackState(state: Object): any;
    /**
     * Returns a way to uniquely identify this component type.
     */
    static getTypeName(): string;
    /**
     * Creates an object that can be used to recreate this component.
     */
    deflate(): DeflatedComponent;
    /**
     * Creates an instance of a component from state generated by $deflate.
     *
     * @param deflated The deflated component state to use
     * @param inst An optional existing object to inflate
     */
    static inflate(deflated: DeflatedComponent, inst?: Component<any>): Component<any>;
    /**
     * Returns an identifier that should be unique among a component's siblings.
     *
     * @param label The label of the component to create the label for
     */
    static getIdentifier(label: string): string;
    /**
     * Returns an identifier that should be unique among a component's siblings.
     */
    getIdentifier(): string;
    toString(): string;
    /**
     * Returns the complete HTML used to render this component.
     */
    getHTML(): string;
    /**
     * Creates an HTML element for this component if one does not exist.
     */
    ensureElement(): void;
    /**
     * Attaches this component to an existing HTML element.
     *
     * @param element The element to attach to
     */
    attachTo(element: HTMLElement): void;
    /**
     * Attempts to locate this component's element from a parent element.
     *
     * @param parent The element to search in
     */
    locate(parent?: HTMLElement): void;
    /**
     * Evaluates the component's template and sets the component's element's HTML to the result.
     */
    render(): void;
    shouldInstall(): boolean;
    /**
     * Attaches event handlers for the component and its children.
     * Often used after render, or as part of reify.
     */
    install(): void;
    installed(): void;
    shouldUninstall(): boolean;
    /**
     * Detaches event handlers for the component and its children.
     */
    uninstall(): void;
    uninstalled(): void;
    /**
     * Ensures this component's element exists, then renders and installs it.
     */
    reify(): void;
}
